#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

from compose_x_common.compose_x_common import keyisset
from troposphere import GetAtt, Ref

from ecs_composex.common import build_template
from ecs_composex.common.compose_resources import XResource
from ecs_composex.common.stacks import ComposeXStack
from ecs_composex.sns.sns_aws import lookup_topic_config
from ecs_composex.sns.sns_params import MOD_KEY, RES_KEY, TOPIC_ARN, TOPIC_NAME
from ecs_composex.sns.sns_perms import ACCESS_TYPES
from ecs_composex.sns.sns_templates import generate_sns_templates
from ecs_composex.sqs.sqs_params import RES_KEY as SQS_KEY


def create_sns_mappings(resources, settings):
    if not keyisset(RES_KEY, settings.mappings):
        mappings = {}
        settings.mappings[RES_KEY] = mappings
    else:
        mappings = settings.mappings[RES_KEY]
    for resource in resources:
        resource_config = lookup_topic_config(
            resource.logical_name, resource.lookup, settings.session
        )
        if resource_config:
            mappings.update({resource.logical_name: resource_config})


class Topic(XResource):
    """
    Class for SNS Topics
    """

    policies_scaffolds = ACCESS_TYPES
    keyword = "Topics"

    def init_outputs(self):
        self.output_properties = {
            TOPIC_ARN: (self.logical_name, self.cfn_resource, Ref, None),
            TOPIC_NAME: (
                f"{self.logical_name}{TOPIC_NAME.title}",
                self.cfn_resource,
                GetAtt,
                TOPIC_NAME.return_value,
            ),
        }


class Subscription(XResource):
    """
    Class for SNS Subscriptions
    """

    keyword = "Subscriptions"


class XStack(ComposeXStack):
    """
    Class to handle SQS Root stack related actions
    """

    def __init__(self, title, settings, **kwargs):
        topics = []
        subscriptions = []
        if keyisset(Topic.keyword, settings.compose_content[RES_KEY]):
            for resource_name in settings.compose_content[RES_KEY][Topic.keyword]:
                topic = Topic(
                    resource_name,
                    settings.compose_content[RES_KEY][Topic.keyword][resource_name],
                    MOD_KEY,
                    settings,
                )
                settings.compose_content[RES_KEY][Topic.keyword][resource_name] = topic
                topics.append(topic)

        if keyisset(Subscription.keyword, settings.compose_content[RES_KEY]):
            for resource_name in settings.compose_content[RES_KEY][
                Subscription.keyword
            ]:
                subscription = Subscription(
                    resource_name,
                    settings.compose_content[RES_KEY][Subscription.keyword][
                        resource_name
                    ],
                    MOD_KEY,
                    settings,
                )
                settings.compose_content[RES_KEY][Subscription.keyword][
                    resource_name
                ] = subscription
                subscriptions.append(subscription)

        new_topics = [topic for topic in topics if not topic.lookup and not topic.use]
        new_subscriptions = [
            subscription
            for subscription in subscriptions
            if not subscription.lookup and not subscription.use
        ]
        if not new_topics and not new_subscriptions:
            self.is_void = True
        else:
            template = build_template(
                "Root template for SNS generated by ECS Compose-X"
            )
            generate_sns_templates(
                settings, new_topics, new_subscriptions, self, template
            )
            super().__init__(title, stack_template=template, **kwargs)
        for topic in topics:
            topic.stack = self
        for subscription in subscriptions:
            subscription.stack = self
        lookup_topics = [
            topic for topic in topics if topic.lookup and not topic.properties
        ]
        if lookup_topics:
            create_sns_mappings(lookup_topics, settings)

    def handle_sqs(self, root_template, sqs_root_stack):
        """
        Function to handle the SQS configuration to allow SNS to send messages to queues.
        """

    def add_xdependencies(self, root_stack, settings):
        """
        Method to add a dependencies from other X-Resources
        :param ComposeXStack root_stack: The ComposeX Root template
        :param ecs_composex.common.ComposeXSettings settings:
        """
        resources = root_stack.stack_template.resources
        sqs_res = SQS_KEY.strip("x-") if SQS_KEY.startswith("x-") else SQS_KEY
        if SQS_KEY in settings.compose_content and sqs_res in resources:
            self.DependsOn.append(sqs_res)
            self.handle_sqs(root_stack.stack_template, resources[sqs_res])



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

"""
Module to manage IAM policies to grant access to ECS Services to DynamodbTables
"""

from ecs_composex.dynamodb.dynamodb_aws import lookup_dynamodb_config
from ecs_composex.dynamodb.dynamodb_params import TABLE_ARN, TABLE_NAME
from ecs_composex.resource_settings import (
    handle_lookup_resource,
    handle_resource_to_services,
)


def create_dyndb_mappings(mapping, resources, settings):
    """
    Function to create the resource mapping for SQS Queues.

    :param dict mapping:
    :param list resources:
    :param ecs_composex.common.settings.ComposeXSettings settings:
    :return:
    """
    for res in resources:
        res_config = lookup_dynamodb_config(res.lookup, settings.session)
        mapping.update({res.logical_name: res_config})


def dynamodb_to_ecs(resources, services_stack, res_root_stack, settings):
    """
    Function to apply SQS settings to ECS Services
    :return:
    """
    resources_mappings = {}
    new_resources = [
        resources[res_name] for res_name in resources if not resources[res_name].lookup
    ]
    lookup_resources = [
        resources[res_name]
        for res_name in resources
        if resources[res_name].lookup and not resources[res_name].properties
    ]
    create_dyndb_mappings(resources_mappings, lookup_resources, settings)
    for new_res in new_resources:
        handle_resource_to_services(
            new_res,
            services_stack,
            res_root_stack,
            settings,
            TABLE_ARN,
            [TABLE_NAME],
            nested=False,
        )
    for resource in lookup_resources:
        handle_lookup_resource(
            resources_mappings, "dynamodb", resource, TABLE_ARN, [TABLE_NAME]
        )



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

import re

from boto3.session import Session
from compose_x_common.compose_x_common import keyisset

from ecs_composex.common import LOG
from ecs_composex.common.aws import (
    define_lookup_role_from_info,
    find_aws_resource_arn_from_tags_api,
)
from ecs_composex.vpc.vpc_params import (
    APP_SUBNETS,
    PUBLIC_SUBNETS,
    STORAGE_SUBNETS,
    VPC_ID,
)

TAGS_KEY = "Tags"


def delete_subnet_from_settings(subnets, subnet_key, vpc_settings):
    """
    Deletes subnets that are not part of the VPC from vpc_settings

    :param list[dict] subnets:
    :param str subnet_key:
    :param dict vpc_settings:
    """
    for subnet_def in subnets:
        if subnet_def["VpcId"] != vpc_settings[VPC_ID.title]:
            for count, subnet_id in enumerate(vpc_settings[subnet_key]):
                if subnet_id == subnet_def["SubnetId"]:
                    LOG.error(
                        f"x-vpc.Lookup - {vpc_settings[subnet_key][count]}"
                        f" is not part of VPC {vpc_settings[VPC_ID.title]}"
                        "Removing it"
                    )
                    vpc_settings[subnet_key].pop(count)


def validate_subnets_belong_with_vpc(vpc_settings, subnet_keys, session=None):
    """
    Function to ensure all subnets belong to the identified VPC

    :param dict vpc_settings:
    :param list[str] subnet_keys:
    :param boto3.session.Session session:
    :raises: boto3.client.exceptions

    """
    if session is None:
        session = Session()
    client = session.client("ec2")
    for subnet_key in subnet_keys:
        subnets_r = client.describe_subnets(
            Filters=[
                {
                    "Name": "vpc-id",
                    "Values": [
                        vpc_settings[VPC_ID.title],
                    ],
                },
            ],
            SubnetIds=vpc_settings[subnet_key],
        )
        if keyisset("Subnets", subnets_r):
            delete_subnet_from_settings(subnets_r["Subnets"], subnet_key, vpc_settings)
        else:
            raise LookupError(
                f"None of the {subnet_key} subnets",
                ",".join(vpc_settings[subnet_key]),
                "are in VPC",
                vpc_settings[VPC_ID.title],
            )
    for key in vpc_settings.keys():
        if not keyisset(key, vpc_settings) and key in subnet_keys:
            raise KeyError(
                f"No subnets for {key} "
                f"have been identified in {vpc_settings[VPC_ID.title]}"
            )


def lookup_x_vpc_settings(lookup, session):
    """
    Method to set VPC settings from x-vpc

    :param boto3.session.Session session:
    :param dict lookup:
    :return: vpc_settings
    :rtype: dict
    """
    vpc_type = "ec2:vpc"
    subnet_type = "ec2:subnet"
    required_keys = [
        VPC_ID.title,
        PUBLIC_SUBNETS.title,
        APP_SUBNETS.title,
        STORAGE_SUBNETS.title,
    ]
    subnets_keys = [
        PUBLIC_SUBNETS.title,
        APP_SUBNETS.title,
        STORAGE_SUBNETS.title,
    ]
    if not all(key in lookup.keys() for key in required_keys):
        raise KeyError(
            "Missing keys for x-vpc Lookup. Got",
            lookup.keys(),
            "Expected",
            required_keys,
        )
    lookup_session = define_lookup_role_from_info(lookup, session)
    vpc_types = {
        vpc_type: {
            "regexp": r"(?:^arn:aws(?:-[a-z]+)?:ec2:[a-z0-9-]+:[0-9]{12}:vpc/)(vpc-[a-z0-9]+)$"
        },
        subnet_type: {
            "regexp": r"(?:^arn:aws(?:-[a-z]+)?:ec2:[a-z0-9-]+:[0-9]{12}:subnet/)(subnet-[a-z0-9]+)$"
        },
    }
    vpc_arn = find_aws_resource_arn_from_tags_api(
        lookup[VPC_ID.title],
        lookup_session,
        vpc_type,
        types=vpc_types,
        allow_multi=False,
    )
    vpc_re = re.compile(vpc_types[vpc_type]["regexp"])
    vpc_settings = {
        VPC_ID.title: vpc_re.match(vpc_arn).groups()[0],
        APP_SUBNETS.title: [],
        STORAGE_SUBNETS.title: [],
        PUBLIC_SUBNETS.title: [],
    }

    for subnet_key in subnets_keys:
        subnet_arns = find_aws_resource_arn_from_tags_api(
            lookup[subnet_key],
            lookup_session,
            subnet_type,
            types=vpc_types,
            allow_multi=True,
        )
        vpc_settings[subnet_key] = [
            re.match(vpc_types[subnet_type]["regexp"], subnet_arn).groups()[0]
            for subnet_arn in subnet_arns
        ]
    extra_subnets = [
        key
        for key in lookup.keys()
        if key not in required_keys and not key == "RoleArn"
    ]
    for subnet_name in extra_subnets:
        subnet_arns = find_aws_resource_arn_from_tags_api(
            lookup[subnet_name],
            lookup_session,
            subnet_type,
            types=vpc_types,
            allow_multi=True,
        )
        vpc_settings[subnet_name] = [
            re.match(vpc_types[subnet_type]["regexp"], subnet_arn).groups()[0]
            for subnet_arn in subnet_arns
        ]
    vpc_settings["session"] = lookup_session
    total_subnets_keys = subnets_keys + extra_subnets
    validate_subnets_belong_with_vpc(
        vpc_settings=vpc_settings,
        subnet_keys=total_subnets_keys,
        session=lookup_session,
    )
    return vpc_settings



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

"""
Module to do a better env variables handling.
"""

import os
import re

ENV_VAR_REGEXP = r"(?<!\\)\$(\w+|\{(?!AWS::)([^}]*)\})"
SPECIAL_INTERPOLATION = r"(?<!\\)(\$(\{(((?!AWS::)[^}]+)(\:[+-=]{1}))([^}]+)\}))"
IF_UNDEFINED = r":-"
IF_DEFINED = r":+"


def expandvars(path, default=None, skip_escaped=True):
    """
    Expand environment variables of form $var and ${var}.
       If parameter 'skip_escaped' is True, all escaped variable references
       (i.e. preceded by backslashes) are skipped.
       Unknown variables are set to 'default'. If 'default' is None,
       they are left unchanged.
    """

    def replace_var(match):
        if re.match(SPECIAL_INTERPOLATION, match.group(0)):
            groups = re.findall(SPECIAL_INTERPOLATION, match.group(0))
            if groups[0][-2] == IF_UNDEFINED:
                return os.environ.get(groups[0][-3]) or expandvars(
                    groups[0][-1], default, skip_escaped
                )
            elif groups[0][-2] == IF_DEFINED:
                return expandvars(groups[0][-1])
        return os.environ.get(
            match.group(2) or match.group(1),
            match.group(0) if default is None else default,
        )

    re_string = (r"(?<!\\)" if skip_escaped else "") + r"\$(\w+|\{(?!AWS::)([^}]*)\})"
    return re.sub(re_string, replace_var, path)



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

"""
Module for VpcStack
"""

import re

from compose_x_common.compose_x_common import keyisset
from troposphere import FindInMap

from ecs_composex.common import LOG, build_template
from ecs_composex.common.cfn_params import Parameter
from ecs_composex.common.ecs_composex import X_KEY
from ecs_composex.common.stacks import ComposeXStack
from ecs_composex.dns import dns_params
from ecs_composex.vpc import aws_mappings
from ecs_composex.vpc.vpc_aws import lookup_x_vpc_settings
from ecs_composex.vpc.vpc_maths import get_subnet_layers
from ecs_composex.vpc.vpc_params import (
    APP_SUBNETS,
    DEFAULT_VPC_CIDR,
    PUBLIC_SUBNETS,
    RES_KEY,
    STORAGE_SUBNETS,
    SUBNETS_TYPE,
    VPC_CIDR,
    VPC_ID,
    VPC_SINGLE_NAT,
)
from ecs_composex.vpc.vpc_subnets import (
    add_apps_subnets,
    add_public_subnets,
    add_storage_subnets,
)
from ecs_composex.vpc.vpc_template import (
    add_template_outputs,
    add_vpc_cidrs_outputs,
    add_vpc_core,
    add_vpc_flow,
)

AZ_INDEX_PATTERN = r"(([a-z0-9-]+)([a-z]{1}$))"
AZ_INDEX_RE = re.compile(AZ_INDEX_PATTERN)


class VpcStack(ComposeXStack):
    """
    Class to create the VPC Stack
    """

    def __init__(self, title, settings, vpc_settings, **kwargs):

        if not keyisset("Endpoints", vpc_settings):
            endpoints = []
        else:
            endpoints = vpc_settings["Endpoints"]

        if endpoints is None:
            endpoints = []
        curated_azs = []
        for az in settings.aws_azs:
            if isinstance(az, dict):
                curated_azs.append(az["ZoneName"])
            elif isinstance(az, str):
                curated_azs.append(az)
        azs_index = [AZ_INDEX_RE.match(az).groups()[-1] for az in curated_azs]
        layers = get_subnet_layers(vpc_settings[VPC_CIDR.title], len(curated_azs))
        template = build_template(
            "VpcTemplate generated via ECS ComposeX",
            [dns_params.PRIVATE_DNS_ZONE_NAME],
        )
        LOG.debug(azs_index)
        template.add_mapping("AwsLbAccounts", aws_mappings.AWS_LB_ACCOUNTS)
        vpc_core = add_vpc_core(template, vpc_settings[VPC_CIDR.title])
        self.vpc = vpc_core[0]
        storage_subnets = add_storage_subnets(template, self.vpc, azs_index, layers)
        public_subnets = add_public_subnets(
            template,
            self.vpc,
            azs_index,
            layers,
            vpc_core[-1],
            vpc_settings[VPC_SINGLE_NAT.title],
        )
        app_subnets = add_apps_subnets(
            template, self.vpc, azs_index, layers, public_subnets[-1], endpoints
        )
        add_template_outputs(
            template,
            self.vpc,
            storage_subnets[1],
            public_subnets[1],
            app_subnets[1],
        )
        if keyisset("EnableFlowLogs", vpc_settings):
            add_vpc_flow(
                template,
                self.vpc,
                boundary=vpc_settings["FlowLogsRoleBoundary"]
                if keyisset("FlowLogsRoleBoundary", vpc_settings)
                else None,
            )
        add_vpc_cidrs_outputs(template, self.vpc, layers)
        super().__init__(title, stack_template=template, **kwargs)


def define_create_settings(create_def):
    """
    Function to create the VPC creation settings

    :param dict create_def:
    :return:
    """
    create_settings = {
        VPC_CIDR.title: create_def[VPC_CIDR.title]
        if keyisset(VPC_CIDR.title, create_def)
        else DEFAULT_VPC_CIDR,
        VPC_SINGLE_NAT.title: True
        if not keyisset(VPC_SINGLE_NAT.title, create_def)
        else create_def[VPC_SINGLE_NAT.title],
        "Endpoints": create_def["Endpoints"]
        if keyisset("Endpoints", create_def)
        else [],
    }
    create_def.update(create_settings)
    return create_def


def create_new_vpc(vpc_xkey, settings, default=False):
    if not default:
        create_settings = define_create_settings(
            settings.compose_content[vpc_xkey]["Create"]
        )
    else:
        create_settings = {
            VPC_CIDR.title: DEFAULT_VPC_CIDR,
            VPC_SINGLE_NAT.title: True,
            "Endpoints": {
                "AwsServices": [{"service": "ecr.dkr"}, {"service": "ecr.api"}]
            },
        }
    vpc_stack = VpcStack(RES_KEY, settings, create_settings)
    vpc_stack.add_parameter(
        {
            dns_params.PRIVATE_DNS_ZONE_NAME.title: FindInMap(
                "Dns",
                "PrivateNamespace",
                dns_params.PRIVATE_DNS_ZONE_NAME.title,
            )
        }
    )
    return vpc_stack


def set_subnets_from_use(subnets_list, vpc_settings, subnets_def):
    for subnet_name in subnets_list:
        if not isinstance(vpc_settings[subnet_name], (list, str)):
            raise TypeError(
                "The subnet_name must be of type",
                str,
                list,
                "Got",
                type(subnet_name),
            )
        subnets = (
            vpc_settings[subnet_name].split(",")
            if isinstance(vpc_settings[subnet_name], str)
            else vpc_settings[subnet_name]
        )
        subnets_def[subnet_name] = subnets


def import_vpc_settings(vpc_settings):
    """
    Function to import settings set "in-stone" from docker-compose definition

    :param dict vpc_settings:
    :return: settings
    :rtype dict:
    """
    if not keyisset(VPC_ID.title, vpc_settings):
        raise KeyError("You must specify the VPC ID to use for deployment")
    settings = {VPC_ID.title: vpc_settings[VPC_ID.title]}
    required_subnets = [
        APP_SUBNETS.title,
        PUBLIC_SUBNETS.title,
        STORAGE_SUBNETS.title,
    ]
    excluded = [VPC_ID.title, "RoleArn"]
    if not all(subnet in vpc_settings.keys() for subnet in required_subnets):
        raise KeyError("All required subnets must be indicated", required_subnets)
    extra_subnets = [
        key
        for key in vpc_settings.keys()
        if key not in required_subnets and key not in excluded
    ]
    set_subnets_from_use(required_subnets, vpc_settings, settings)
    set_subnets_from_use(extra_subnets, vpc_settings, settings)

    return settings


def create_vpc_mapping(settings_params):
    """
    Function to create a CFN Mapping to use and assign subnets to substacks

    :param settings_params:
    :return:
    """


def apply_vpc_settings(x_settings, root_stack, settings):
    """

    :param x_settings:
    :param root_stack:
    :param ecs_composex.common.settings.ComposeXSettings settings:
    :return:
    """
    settings.subnets_mappings = {
        VPC_ID.title: {VPC_ID.title: x_settings[VPC_ID.title]},
        APP_SUBNETS.title: {"Ids": x_settings[APP_SUBNETS.title]},
        STORAGE_SUBNETS.title: {"Ids": x_settings[STORAGE_SUBNETS.title]},
        PUBLIC_SUBNETS.title: {"Ids": x_settings[PUBLIC_SUBNETS.title]},
    }
    ignored_keys = ["RoleArn", "session"]
    settings.subnets_parameters.append(APP_SUBNETS)
    settings.subnets_parameters.append(PUBLIC_SUBNETS)
    settings.subnets_parameters.append(STORAGE_SUBNETS)
    for setting_name in x_settings:
        if (
            setting_name not in settings.subnets_mappings.keys()
            and setting_name not in ignored_keys
        ):
            settings.subnets_mappings[setting_name] = {"Ids": x_settings[setting_name]}
            param = Parameter(setting_name, Type=SUBNETS_TYPE)
            settings.subnets_parameters.append(param)

    root_stack.stack_template.add_mapping("Network", settings.subnets_mappings)
    settings.set_azs_from_vpc_import(
        x_settings,
        session=x_settings["session"] if keyisset("session", x_settings) else None,
    )


def add_vpc_to_root(root_stack, settings):
    """
    Function to figure whether to create the VPC Stack and if not, set the parameters.

    :param root_stack:
    :param settings:
    :return: vpc_stack
    :rtype: VpcStack
    """
    vpc_stack = None
    vpc_xkey = f"{X_KEY}{RES_KEY}"

    if keyisset(vpc_xkey, settings.compose_content):
        if keyisset("Lookup", settings.compose_content[vpc_xkey]):
            settings.create_vpc = False
            x_settings = lookup_x_vpc_settings(
                settings.compose_content[vpc_xkey]["Lookup"], settings.session
            )
            apply_vpc_settings(x_settings, root_stack, settings)
        elif keyisset("Use", settings.compose_content[vpc_xkey]):
            x_settings = import_vpc_settings(settings.compose_content[vpc_xkey]["Use"])
            apply_vpc_settings(x_settings, root_stack, settings)
        else:
            if keyisset("Create", settings.compose_content[vpc_xkey]) and keyisset(
                "Lookup", settings.compose_content[vpc_xkey]
            ):
                settings.create_vpc = True
                LOG.warning(
                    "We have both Create and Lookup set for x-vpc. Creating a new VPC"
                )
            vpc_stack = create_new_vpc(vpc_xkey, settings)
            settings.create_vpc = True
    else:
        LOG.info(f"No {vpc_xkey} detected. Creating a new VPC.")
        vpc_stack = create_new_vpc(vpc_xkey, settings, default=True)
        settings.create_vpc = True
    if isinstance(vpc_stack, VpcStack):
        root_stack.stack_template.add_resource(vpc_stack)
    return vpc_stack



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

"""
Core ECS Template building
"""

from troposphere import GetAtt, Ref, Sub, Tags
from troposphere.ec2 import SecurityGroup
from troposphere.iam import Policy
from troposphere.logs import LogGroup

from ecs_composex.common import build_template
from ecs_composex.common.cfn_params import ROOT_STACK_NAME, ROOT_STACK_NAME_T
from ecs_composex.dns import dns_conditions, dns_params
from ecs_composex.dns.dns_conditions import (
    CREATE_PUBLIC_NAMESPACE_CON,
    CREATE_PUBLIC_NAMESPACE_CON_T,
)
from ecs_composex.ecs import ecs_conditions, ecs_params
from ecs_composex.ecs.ecs_params import CLUSTER_NAME, CLUSTER_NAME_T
from ecs_composex.ecs.ecs_service import Service
from ecs_composex.ecs.ecs_service_config import ServiceConfig
from ecs_composex.secrets.secrets_params import RES_KEY as SECRETS_KEY
from ecs_composex.vpc import vpc_params


def initialize_service_template(service_name):
    """Function to initialize the base template for ECS Services with all
    parameters and conditions necessary for CFN to work properly

    :param service_name: Name of the ecs_service as defined in ComposeX File
    :type service_name: str

    :return: service_template
    :rtype: troposphere.Template
    """
    service_tpl = build_template(
        f"Template for {service_name}",
        [
            dns_params.PUBLIC_DNS_ZONE_NAME,
            dns_params.PRIVATE_DNS_ZONE_NAME,
            dns_params.PUBLIC_DNS_ZONE_ID,
            dns_params.PRIVATE_DNS_ZONE_ID,
            dns_params.PRIVATE_NAMESPACE_ID,
            ecs_params.CLUSTER_NAME,
            ecs_params.LAUNCH_TYPE,
            ecs_params.ECS_CONTROLLER,
            ecs_params.SERVICE_COUNT,
            ecs_params.CLUSTER_SG_ID,
            ecs_params.SERVICE_HOSTNAME,
            ecs_params.FARGATE_CPU_RAM_CONFIG,
            ecs_params.SERVICE_NAME,
            ecs_params.ELB_GRACE_PERIOD,
            ecs_params.FARGATE_VERSION,
            ecs_params.LOG_GROUP_RETENTION,
            vpc_params.VPC_ID,
            vpc_params.APP_SUBNETS,
            vpc_params.PUBLIC_SUBNETS,
        ],
    )
    service_tpl.add_condition(
        ecs_conditions.SERVICE_COUNT_ZERO_CON_T,
        ecs_conditions.SERVICE_COUNT_ZERO_CON,
    )
    service_tpl.add_condition(
        ecs_conditions.SERVICE_COUNT_ZERO_AND_FARGATE_CON_T,
        ecs_conditions.SERVICE_COUNT_ZERO_AND_FARGATE_CON,
    )
    service_tpl.add_condition(
        ecs_conditions.USE_HOSTNAME_CON_T, ecs_conditions.USE_HOSTNAME_CON
    )
    service_tpl.add_condition(
        ecs_conditions.NOT_USE_HOSTNAME_CON_T,
        ecs_conditions.NOT_USE_HOSTNAME_CON,
    )
    service_tpl.add_condition(
        ecs_conditions.NOT_USE_CLUSTER_SG_CON_T,
        ecs_conditions.NOT_USE_CLUSTER_SG_CON,
    )
    service_tpl.add_condition(
        ecs_conditions.USE_CLUSTER_SG_CON_T, ecs_conditions.USE_CLUSTER_SG_CON
    )
    service_tpl.add_condition(
        ecs_conditions.USE_FARGATE_CON_T,
        ecs_conditions.USE_FARGATE_CON,
    )
    service_tpl.add_condition(
        ecs_conditions.USE_CLUSTER_CAPACITY_PROVIDERS_CON_T,
        ecs_conditions.USE_CLUSTER_CAPACITY_PROVIDERS_CON,
    )
    service_tpl.add_condition(
        CREATE_PUBLIC_NAMESPACE_CON_T, CREATE_PUBLIC_NAMESPACE_CON
    )
    service_tpl.add_condition(
        dns_conditions.PRIVATE_ZONE_ID_CON_T, dns_conditions.PRIVATE_ZONE_ID_CON
    )
    service_tpl.add_condition(
        dns_conditions.PRIVATE_NAMESPACE_CON_T,
        dns_conditions.PRIVATE_NAMESPACE_CON,
    )
    return service_tpl


def create_log_group(family):
    """
    Function to create a new Log Group for the services
    :return:
    """
    svc_log = family.template.add_resource(
        LogGroup(
            ecs_params.LOG_GROUP_T,
            RetentionInDays=Ref(ecs_params.LOG_GROUP_RETENTION),
            LogGroupName=Sub(
                f"${{{ROOT_STACK_NAME.title}}}/"
                f"svc/ecs/${{{ecs_params.CLUSTER_NAME_T}}}/{family.logical_name}",
            ),
        ),
    )
    family.exec_role.Policies.append(
        Policy(
            f"{family.logical_name}LogGroupAccess",
            PolicyName=Sub(f"CloudWatchAccessForFamily{family.logical_name}"),
            PolicyDocument={
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Sid": "AllowCloudWatchLoggingToSpecificLogGroup",
                        "Effect": "Allow",
                        "Action": ["logs:CreateLogStream", "logs:PutLogEvents"],
                        "Resource": [GetAtt(svc_log, "Arn")],
                    }
                ],
            },
        )
    )


def add_clusterwide_security_group(template):
    """
    Function to generate the ecs_service Load Balancers (if Any)
    """
    sg = SecurityGroup(
        "ClusterWideSecurityGroup",
        template=template,
        GroupDescription=Sub(f"SG for ${{{CLUSTER_NAME_T}}}"),
        GroupName=Sub(f"cluster-${{{CLUSTER_NAME_T}}}"),
        Tags=Tags(
            {
                "Name": Sub(f"clustersg-${{{CLUSTER_NAME_T}}}"),
                "ClusterName": Ref(CLUSTER_NAME),
            }
        ),
        VpcId=Ref(vpc_params.VPC_ID),
    )
    return sg


def get_service_family_name(services_families, service_name):
    """
    Function to return the root family name, representing the service stack name.

    :param services_families:
    :param service_name:
    :return: service stack name
    :rtype: str
    """
    for family_name in services_families:
        if service_name in services_families[family_name]:
            return family_name
    if service_name in services_families.keys():
        return service_name
    return None


def generate_services(settings):
    """
    Function to handle creation of services within the same family.

    :param ecs_composex.common.settings.ComposeXSettings settings:
    :return:
    """
    for family_name in settings.families:
        family = settings.families[family_name]
        family.template = initialize_service_template(family_name)
        if settings.secrets_mappings:
            family.template.add_mapping(SECRETS_KEY, settings.secrets_mappings)
        family.init_task_definition()
        family.set_secrets_access()
        family.refresh()
        family.assign_policies()
        family.service_config = ServiceConfig(family, settings)
        family.ecs_service = Service(family, settings)
        family.service_config.network.set_aws_sources(
            settings,
            family.logical_name,
            GetAtt(family.ecs_service.sg, "GroupId"),
        )
        family.service_config.network.set_ext_sources_ingress(
            family.logical_name, GetAtt(family.ecs_service.sg, "GroupId")
        )
        family.service_config.network.associate_aws_igress_rules(family.template)
        family.service_config.network.associate_ext_igress_rules(family.template)
        family.service_config.network.add_self_ingress(family)
        family.stack_parameters.update(
            {
                ecs_params.SERVICE_NAME_T: family.logical_name,
                CLUSTER_NAME_T: Ref(CLUSTER_NAME),
                ROOT_STACK_NAME_T: Ref(ROOT_STACK_NAME),
                ecs_params.LAUNCH_TYPE.title: family.compute_platform,
            }
        )
        family.upload_services_env_files(settings)
        family.set_repository_credentials(settings)
        family.set_volumes()
        create_log_group(family)
        family.handle_logging()
        family.handle_alarms()
        family.handle_prometheus()



#   -*- coding: utf-8 -*-
#  SPDX-License-Identifier: MPL-2.0
#  Copyright 2020-2021 John Mille <john@compose-x.io>

import re
from os import path

from ecs_composex.common.compose_resources import Parameter
from ecs_composex.common.ecs_composex import X_KEY

MOD_KEY = path.basename(path.dirname(path.abspath(__file__)))
RES_KEY = f"{X_KEY}{MOD_KEY}"
MAPPINGS_KEY = re.sub(r"[^a-zA-Z0-9]", "", MOD_KEY)

USERPOOL_ID = Parameter("UserPoolId", Type="String")
USERPOOL_ARN = Parameter("UserPoolArn", return_value="Arn", Type="String")
USERPOOL_PROVIDER_NAME = Parameter(
    "UserPoolProviderName", return_value="ProviderName", Type="String"
)
USERPOOL_PROVIDER_URL = Parameter(
    "UserPoolProviderUrl", return_value="ProviderURL", Type="String"
)
USERPOOL_DOMAIN = Parameter("UserPoolDomain", Type="String")



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

"""
Module to create the ElasticCache Cluster and nodes
"""

from compose_x_common.compose_x_common import keyisset
from troposphere import AWS_STACK_NAME, GetAtt, Ref, Sub, Tags
from troposphere.ec2 import SecurityGroup
from troposphere.elasticache import (
    CacheCluster,
    ParameterGroup,
    ReplicationGroup,
    SubnetGroup,
)

from ecs_composex.common import LOG, build_template
from ecs_composex.resources_import import import_record_properties
from ecs_composex.vpc.vpc_params import STORAGE_SUBNETS, VPC_ID


def init_root_template():
    return build_template("Root stack for ElasticCache", [VPC_ID, STORAGE_SUBNETS])


def create_replication_group(cluster):
    """
    Function to add the replication group from properties

    :param ecs_composex.elasticache.elasticache_stack.CacheCluster cluster:
    :return:
    """
    if (
        not cluster.cfn_resource
        and isinstance(cluster.cfn_resource, CacheCluster)
        and cluster.cfn_resource.Engine == "redis"
    ):
        raise ValueError("Replication group can only be set for a Redis cache cluster.")

    to_del = ["CacheSecurityGroupNames", "PreferredCacheClusterAZs"]
    props = import_record_properties(
        cluster.parameters["ReplicationGroup"], ReplicationGroup
    )
    for prop_to_del in to_del:
        if keyisset(prop_to_del, props):
            del props[prop_to_del]
    if not (
        keyisset("NumCacheClusters", props)
        or keyisset("NumNodeGroups", props)
        or keyisset("ReplicasPerNodeGroup", props)
    ):
        props["PrimaryClusterId"] = Ref(cluster.cfn_resource)
    props["SecurityGroupIds"] = [GetAtt(cluster.db_sg, "GroupId")]
    if cluster.parameter_group:
        props["CacheParameterGroupName"] = Ref(cluster.parameter_group)
    group = ReplicationGroup(f"{cluster.logical_name}ReplicationGroup", **props)
    return group


def create_parameter_group(cluster, definition):
    """
    Function to add the parameter group

    :param ecs_composex.elasticache.elasticache_stack.CacheCluster cluster:
    :param dict definition:
    :return: The parameter group
    :rtype: troposphere.elasticache.ParameterGroup
    """
    props = import_record_properties(definition, ParameterGroup)
    cluster.parameter_group = ParameterGroup(
        f"{cluster.logical_name}ParameterGroup", **props
    )


def determine_resource_type(name, properties):
    """
    Function to determine if the properties are the ones of a DB Cluster or DB Instance.
    By default it will assume Cluster if cannot conclude that it is a DB Instance

    :param str name:
    :param dict properties:
    :return:
    """
    if all(
        property_name in CacheCluster.props.keys()
        for property_name in properties.keys()
    ):
        LOG.info(f"Identified {name} to be {CacheCluster.resource_type}")
        return CacheCluster
    elif all(
        property_name in ReplicationGroup.props.keys()
        for property_name in properties.keys()
    ):
        LOG.info(f"Identified {name} to be {ReplicationGroup.resource_type}")
        return ReplicationGroup
    LOG.error(
        "From the properties defined, we cannot determine whether this is a RDS Cluster or RDS Instance."
        " Setting to Cluster"
    )
    return None


def handle_security_groups(cluster, props, resource_class):
    """
    Function to handle security groups properties and assignment.

    :param ecs_composex.elasticache.elasticache_stack.CacheCluster cluster:
    :param dict props:
    :param resource_class:
    :raises: TypeError
    """
    if resource_class is CacheCluster:
        key = "VpcSecurityGroupIds"
    elif resource_class is ReplicationGroup:
        key = "SecurityGroupIds"
    else:
        raise TypeError(
            "resource_class must be one of",
            [CacheCluster, ReplicationGroup],
            "Got",
            type(resource_class),
        )

    if keyisset(key, props) and not isinstance(props[key], list):
        raise TypeError(f"{key} must be a list. Got", type(props[key]))
    elif keyisset(key, props):
        props[key].append(GetAtt(cluster.db_sg, "GroupId"))
    else:
        props[key] = [GetAtt(cluster.db_sg, "GroupId")]


def create_cluster_from_properties(cluster, template):
    """
    Function to create the Elastic Cache Cluster from properties

    :param ecs_composex.elasticache.elasticache_stack.CacheCluster cluster:
    :param troposphere.Template template:
    :return:
    """
    resource_class = determine_resource_type(cluster.name, cluster.properties)
    props = import_record_properties(cluster.properties, resource_class)
    props["CacheSubnetGroupName"] = Ref(cluster.db_subnet_group)
    handle_security_groups(cluster, props, resource_class)
    default_tags = Tags(Name=cluster.logical_name, ComposeName=cluster.name)
    if keyisset("Tags", props):
        props["Tags"] += default_tags
    else:
        props["Tags"] = default_tags
    if cluster.parameters and keyisset("ParameterGroup", cluster.parameters):
        create_parameter_group(cluster, cluster.parameters["ParameterGroup"])
        template.add_resource(cluster.parameter_group)
        props["CacheParameterGroupName"] = Ref(cluster.parameter_group)
    cluster.cfn_resource = resource_class(cluster.logical_name, **props)
    template.add_resource(cluster.cfn_resource)


def create_cluster_from_parameters(cluster, template):
    """
    Function to create the Cluster from the MacroParameters

    :param ecs_composex.elasticache.elasticache_stack.CacheCluster cluster:
    :param template:
    :return:
    """
    required_keys = ["Engine", "EngineVersion"]
    if not cluster.properties and not all(
        key in required_keys for key in cluster.parameters
    ):
        raise KeyError(
            "When using MacroParameters only, you must specify at least",
            required_keys,
        )
    props = {
        "CacheNodeType": "cache.t3.small"
        if not keyisset("CacheNodeType", cluster.parameters)
        else cluster.parameters["CacheNodeType"],
        "Engine": cluster.parameters["Engine"],
        "EngineVersion": cluster.parameters["EngineVersion"],
        "NumCacheNodes": 1,
        "VpcSecurityGroupIds": [GetAtt(cluster.db_sg, "GroupId")],
        "CacheSubnetGroupName": Ref(cluster.db_subnet_group),
        "Tags": Tags(Name=cluster.logical_name, ComposeName=cluster.name),
    }
    if keyisset("ParameterGroup", cluster.parameters):
        create_parameter_group(cluster, cluster.parameters["ParameterGroup"])
        template.add_resource(cluster.parameter_group)
    cluster.cfn_resource = CacheCluster(cluster.logical_name, **props)
    template.add_resource(cluster.cfn_resource)


def create_root_template(new_resources):
    """
    Function to create the root template and add the new resources to it.

    :param list<ecs_composex.elasticache.elasticache_stack.CacheCluster> new_resources:
    :return: the root template for ElasticCache
    :rtype: troposphere.Template
    """

    root_template = init_root_template()
    for resource in new_resources:
        resource.db_subnet_group = SubnetGroup(
            f"{resource.logical_name}SubnetGroup",
            Description="ElasticCacheSubnetGroup",
            SubnetIds=Ref(STORAGE_SUBNETS)
            if not resource.subnets_override
            else Ref(resource.subnets_override),
        )

        resource.db_sg = SecurityGroup(
            f"{resource.logical_name}Sg",
            GroupDescription=Sub(f"SG for docdb-{resource.logical_name}"),
            GroupName=Sub(f"${{{AWS_STACK_NAME}}}.elasticache.{resource.logical_name}"),
            VpcId=Ref(VPC_ID),
        )
        root_template.add_resource(resource.db_sg)
        root_template.add_resource(resource.db_subnet_group)
        if resource.properties:
            create_cluster_from_properties(resource, root_template)
        elif resource.parameters and not resource.properties:
            create_cluster_from_parameters(resource, root_template)

        if isinstance(resource.cfn_resource, CacheCluster):
            if resource.cfn_resource.Engine == "memcached":
                resource.init_memcached_outputs()
                resource.add_memcahed_config(root_template)
            elif resource.cfn_resource.Engine == "redis":
                resource.init_redis_outputs()
                resource.add_redis_config(root_template)
        elif isinstance(resource.cfn_resource, ReplicationGroup):
            resource.init_redis_replica_outputs()
            resource.add_redis_replica_config(root_template)
        resource.generate_outputs()
        root_template.add_output(resource.outputs)
    return root_template



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

"""
Core module for ECS ComposeX.

This module is going to parse each ecs_service and each x-resource key from the compose file
(hence ComposeX) and determine its

* ServiceDefinition
* TaskDefinition
* TaskRole
* ExecutionRole

It is going to also, based on the labels set in the compose file

* Add the ecs_service to Service Discovery via AWS CloudMap
* Add load-balancers to dispatch traffic to the microservice

"""

from ecs_composex import __version__ as version

metadata = {
    "Type": "ComposeX",
    "Properties": {
        "ecs_composex::module": "ecs_composex.ecs",
        "Version": version,
    },
}



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

"""
AWS DocumentDB entrypoint for ECS ComposeX
"""

from troposphere import GetAtt, Ref

from ecs_composex.common.compose_resources import XResource, set_resources
from ecs_composex.common.stacks import ComposeXStack
from ecs_composex.docdb.docdb_params import (
    DOCDB_NAME,
    DOCDB_PORT,
    DOCDB_SECRET,
    DOCDB_SG,
    MOD_KEY,
    RES_KEY,
)
from ecs_composex.docdb.docdb_template import (
    create_docdb_template,
    init_doc_db_template,
)
from ecs_composex.vpc.vpc_params import STORAGE_SUBNETS


class DocDb(XResource):
    """
    Class to manage DocDB
    """

    subnets_param = STORAGE_SUBNETS

    def __init__(self, name, definition, module_name, settings, mapping_key=None):
        """
        Init method

        :param str name:
        :param dict definition:
        :param ecs_composex.common.settings.ComposeXSettings settings:
        """
        self.db_secret = None
        self.db_sg = None
        self.db_subnets_group = None
        super().__init__(
            name, definition, module_name, settings, mapping_key=mapping_key
        )
        self.set_override_subnets()

    def init_outputs(self):
        """
        Method to init the DocDB output attributes
        """
        self.output_properties = {
            DOCDB_NAME: (self.logical_name, self.cfn_resource, Ref, None),
            DOCDB_PORT: (
                f"{self.logical_name}{DOCDB_PORT.return_value}",
                self.cfn_resource,
                GetAtt,
                DOCDB_PORT.return_value,
            ),
            DOCDB_SECRET: (
                self.db_secret.title,
                self.db_secret,
                Ref,
                None,
            ),
            DOCDB_SG: (
                self.db_sg.title,
                self.db_sg,
                GetAtt,
                DOCDB_SG.return_value,
            ),
        }


class XStack(ComposeXStack):
    """
    Class for the Stack of DocDB
    """

    def __init__(self, title, settings, **kwargs):
        set_resources(settings, DocDb, RES_KEY, MOD_KEY)
        new_resources = [
            docdb
            for docdb in settings.compose_content[RES_KEY].values()
            if not docdb.lookup and not docdb.use
        ]
        if new_resources:
            stack_template = init_doc_db_template()
            super().__init__(title, stack_template, **kwargs)
            create_docdb_template(stack_template, new_resources, settings, self)
        else:
            self.is_void = True
        for resource in settings.compose_content[RES_KEY].values():
            resource.stack = self



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

from json import loads
from os import path

from ecs_composex.iam.import_sam_policies import import_and_cleanse_policies


def get_access_types():
    sam_policies = import_and_cleanse_policies()
    with open(
        f"{path.abspath(path.dirname(__file__))}/kinesis_perms.json",
        "r",
        encoding="utf-8-sig",
    ) as perms_fd:
        kinesis_policies = loads(perms_fd.read())
    sam_policies.update(kinesis_policies)
    return sam_policies


ACCESS_TYPES = get_access_types()



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

"""
Module to help with defining the network settings for the ECS Service based on the family services definitions.
"""

import re
from copy import deepcopy
from ipaddress import IPv4Interface
from json import dumps

from compose_x_common.compose_x_common import keyisset
from troposphere import AWS_ACCOUNT_ID, AWS_NO_VALUE, Ref, Sub
from troposphere.ec2 import SecurityGroupIngress

from ecs_composex.common import LOG, NONALPHANUM
from ecs_composex.common.aws import (
    define_lookup_role_from_info,
    find_aws_resource_arn_from_tags_api,
)


def flatten_ip(ip_str):
    """
    Function to remove all non alphanum characters from IP CIDR notation

    :param ip_str:
    :rtype: str
    """
    return str(ip_str.replace(".", "").split("/")[0].strip())


def generate_security_group_props(allowed_source):
    """
    Function to parse the allowed source and create the SG Opening options accordingly.

    :param dict allowed_source: The allowed source defined in configs
    :return: security group ingress properties
    :rtype: dict
    """
    props = {
        "CidrIp": (
            allowed_source[Ingress.ipv4_key]
            if keyisset(Ingress.ipv4_key, allowed_source)
            else Ref(AWS_NO_VALUE)
        ),
        "CidrIpv6": (
            allowed_source[Ingress.ipv6_key]
            if keyisset(Ingress.ipv6_key, allowed_source)
            else Ref(AWS_NO_VALUE)
        ),
    }
    if keyisset("CidrIp", props) and isinstance(props["CidrIp"], str):
        try:
            IPv4Interface(props["CidrIp"])
        except Exception as error:
            LOG.error(f"Falty IP Address: {allowed_source}")
            raise ValueError("Not a valid IPv4 CIDR notation", props["CidrIp"], error)
    return props


def define_protocol(port_string):
    """
    Function to define the port protocol. Defaults to TCP if not specified otherwise

    :param port_string: the port string to parse from the ports list in the compose file
    :type port_string: str
    :return: protocol, ie. udp or tcp
    :rtype: str
    """
    protocols = ["tcp", "udp"]
    protocol = "tcp"
    if port_string.find("/"):
        protocol_found = port_string.split("/")[-1].strip()
        if protocol_found in protocols:
            return protocol_found
    return protocol


def set_service_ports(ports):
    """Function to define common structure to ports

    :return: list of ports the ecs_service uses formatted according to dict
    :rtype: list
    """
    service_ports = []
    for port in ports:
        if not isinstance(port, (str, dict, int)):
            raise TypeError(
                "ports must be of types", dict, "or", list, "got", type(port)
            )
        if isinstance(port, str):
            service_ports.append(
                {
                    "protocol": define_protocol(port),
                    "published": int(port.split(":")[0]),
                    "target": int(port.split(":")[-1].split("/")[0].strip()),
                    "mode": "awsvpc",
                }
            )
        elif isinstance(port, dict):
            valid_keys = ["published", "target", "protocol", "mode"]
            if not set(port).issubset(valid_keys):
                raise KeyError("Valid keys are", valid_keys, "got", port.keys())
            port["mode"] = "awsvpc"
            service_ports.append(port)
        elif isinstance(port, int):
            service_ports.append(
                {
                    "protocol": "tcp",
                    "published": port,
                    "target": port,
                    "mode": "awsvpc",
                }
            )
    LOG.debug(service_ports)
    return service_ports


def lookup_security_group(settings, lookup):
    """
    Function to fetch the security group ID based on lookup details

    :param ecs_composex.common.settings.ComposeXSettings settings:
    :param lookup:
    :return:
    """
    sg_re = re.compile(
        r"(?:^arn:aws(?:-[a-z]+)?:ec2:[a-z0-9-]+:\d{12}:security-group/)([\S]+)$"
    )
    ec2_types = {
        "ec2:security-group": {"regexp": sg_re.pattern},
    }
    lookup_session = define_lookup_role_from_info(lookup, settings.session)
    sg_arn = find_aws_resource_arn_from_tags_api(
        lookup,
        lookup_session,
        "ec2:security-group",
        types=ec2_types,
    )
    if not sg_arn:
        raise LookupError("Failed to identify EC2 SecurityGroup based on tags")
    return sg_re.match(sg_arn).groups()[0]


class Ingress(object):
    """
    Class to group the configuration for Service network settings
    """

    defined = True

    master_key = "Ingress"
    aws_sources_key = "AwsSources"
    ext_sources_key = "ExtSources"
    services_key = "Services"
    ipv4_key = "IPv4"
    ipv6_key = "IPv6"
    network_settings = [master_key, "UseCloudmap", "IsPublic"]

    def __init__(self, definition, ports):
        """
        Initialize network settings for the family ServiceConfig
        """
        self.definition = deepcopy(definition)

        self.aws_sources = (
            self.definition[self.aws_sources_key]
            if keyisset(self.aws_sources_key, self.definition)
            else []
        )
        self.ext_sources = (
            self.definition[self.ext_sources_key]
            if keyisset(self.ext_sources_key, self.definition)
            else []
        )
        self.ext_sources = [
            dict(y) for y in set(tuple(x.items()) for x in self.ext_sources)
        ]
        self.services = (
            self.definition[self.services_key]
            if keyisset(self.services_key, self.definition)
            else []
        )
        self.ports = ports
        self.aws_ingress_rules = []
        self.ext_ingress_rules = []
        self.to_self_rules = []

    def __repr__(self):
        return dumps(self.definition, indent=2)

    def validate_aws_sources(self):
        allowed_keys = ["Type", "Id", "Lookup"]
        allowed_types = ["SecurityGroup", "PrefixList"]
        for source in self.aws_sources:
            if not all(key in allowed_keys for key in source.keys()):
                raise KeyError(
                    "Missing ingress properties. Got",
                    source.keys,
                    "Expected",
                    allowed_keys,
                )
            if not source["Type"] in allowed_types:
                raise ValueError(
                    "Invalid type specified. Got",
                    source["Type"],
                    "Allowed one of ",
                    allowed_types,
                )

    def set_aws_sources(self, settings, destination_title, sg_ref):
        """
        Method to define AWS Sources ingresses

        :param destination_title:
        :param sg_ref:
        :return:
        """
        self.validate_aws_sources()
        for source in self.aws_sources:
            for port in self.ports:
                common_args = {
                    "FromPort": port["published"],
                    "ToPort": port["published"],
                    "IpProtocol": port["protocol"],
                    "GroupId": sg_ref,
                }
                if source["Type"] == "SecurityGroup":
                    if keyisset("Id", source):
                        sg_id = source["Id"]
                    elif keyisset("Lookup", source):
                        sg_id = lookup_security_group(settings, source["Lookup"])
                    else:
                        raise KeyError(
                            "Information missing to identify the SecurityGroup. Requires either Id or Lookup"
                        )
                    common_args.update(
                        {
                            "Description": Sub(
                                f"From {sg_id} to {destination_title} on port {port['published']}"
                            )
                        }
                    )
                    self.aws_ingress_rules.append(
                        SecurityGroupIngress(
                            f"From{NONALPHANUM.sub('', sg_id)}ToServiceOn{port['published']}",
                            SourceSecurityGroupId=sg_id,
                            SourceSecurityGroupOwnerId=Ref(AWS_ACCOUNT_ID),
                            **common_args,
                        )
                    )
                elif source["Type"] == "PrefixList":
                    self.aws_ingress_rules.append(
                        SecurityGroupIngress(
                            f"From{NONALPHANUM.sub('', source['Id'])}ToServiceOn{port['published']}",
                            SourcePrefixListId=source["Id"],
                            **common_args,
                        )
                    )

    def create_ext_sources_ingress_rule(
        self, destination_tile, allowed_source, security_group, **props
    ):
        for port in self.ports:
            if keyisset("Name", allowed_source):
                name = NONALPHANUM.sub("", allowed_source["Name"])
                title = f"From{name.title()}To{port['published']}{port['protocol']}"
                description = Sub(
                    f"From {name.title()} "
                    f"To {port['published']}{port['protocol']} for {destination_tile}"
                )
            else:
                title = (
                    f"From{flatten_ip(allowed_source[self.ipv4_key])}"
                    f"To{port['published']}{port['protocol']}"
                )
                description = Sub(
                    f"Public {port['published']}{port['protocol']}"
                    f" for {destination_tile}"
                )
            self.ext_ingress_rules.append(
                SecurityGroupIngress(
                    title,
                    Description=description
                    if not keyisset("Description", allowed_source)
                    else allowed_source["Description"],
                    GroupId=security_group,
                    IpProtocol=port["protocol"],
                    FromPort=port["published"],
                    ToPort=port["published"],
                    **props,
                )
            )

    def set_ext_sources_ingress(self, destination_tile, security_group):
        """
        Method to add ingress rules from external sources to a given Security Group (ie. ALB Security Group).
        If a list of IPs is found in the config['ext_sources'] part of the network section of configs for the service,
        then it will use that. If no IPv4 source is indicated, it will by default allow traffic from 0.0.0.0/0

        :param str destination_tile: The name of the destination for description
        :param security_group: security group (object or title string) to add the rules to
        :type security_group: str or troposphere.ec2.SecurityGroup or troposphere.Ref or Troposphere.GetAtt
        """
        if not self.ext_sources:
            LOG.info("No external rules defined. Skipping.")
            return

        for allowed_source in self.ext_sources:
            if not keyisset(self.ipv4_key, allowed_source) and not keyisset(
                self.ipv6_key, allowed_source
            ):
                LOG.warning(f"No {self.ipv4_key} or {self.ipv6_key} set. Skipping")
                continue
            props = generate_security_group_props(allowed_source)
            if props:
                LOG.debug(f"Adding {allowed_source} for ingress")
                self.create_ext_sources_ingress_rule(
                    destination_tile, allowed_source, security_group, **props
                )

    def associate_aws_igress_rules(self, template):
        """
        Method to associate AWS ingress rules to a specific template

        :param troposphere.Template template:
        :return:
        """
        for ingress_rule in self.aws_ingress_rules:
            if ingress_rule.title not in template.resources:
                template.add_resource(ingress_rule)

    def associate_ext_igress_rules(self, template):
        """
        Method to associate External ingress rules to a specific template

        :param troposphere.Template template:
        :return:
        """
        for ingress_rule in self.ext_ingress_rules:
            if ingress_rule.title not in template.resources:
                template.add_resource(ingress_rule)



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

from os import path

from ecs_composex.common.cfn_params import Parameter
from ecs_composex.common.ecs_composex import X_KEY
from ecs_composex.vpc.vpc_params import SG_ID_TYPE

MOD_KEY = path.basename(path.dirname(path.abspath(__file__)))
RES_KEY = f"{X_KEY}{MOD_KEY}"


DOCDB_NAME_T = "DocDBName"
DOCDB_ID_T = "ClusterResourceId"
DOCDB_ENDPOINT_T = "Endpoint"
DOCDB_READ_ENDPOINT_T = "ReadEndpoint"
DOCDB_PORT_T = "Port"
DOCDB_SG_T = "DocDbSg"
DOCDB_SECRET_T = "DocDbSecret"
DOCDB_SUBNET_GROUP_T = "DocDbSubnetGroup"

DOCDB_NAME = Parameter(DOCDB_NAME_T, Type="String")
DOCDB_ID_T = Parameter(DOCDB_ID_T, return_value="ClusterResourceId", Type="String")
DOCDBC_ENDPOINT = Parameter(DOCDB_ENDPOINT_T, return_value="Endpoint", Type="String")
DOCDBC_READ_ENDPOINT = Parameter(
    DOCDB_READ_ENDPOINT_T, return_value="ReadEndpoint", Type="String"
)
DOCDB_PORT = Parameter(
    DOCDB_PORT_T,
    return_value="Port",
    Type="Number",
    MinValue=1,
    MaxValue=((2 ^ 16) - 1),
)
DOCDB_SG = Parameter(DOCDB_SG_T, return_value="GroupId", Type=SG_ID_TYPE)
DOCDB_SECRET = Parameter(DOCDB_SECRET_T, Type="String")



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>


from troposphere import MAX_OUTPUTS

from ecs_composex.common import build_template
from ecs_composex.common.stacks import ComposeXStack

CFN_MAX_OUTPUTS = MAX_OUTPUTS - 10


def create_kms_template(template, new_keys, xstack):
    """
    Function to create all the KMS Keys based on their definition

    :param ecs_composex.common.settings.ComposeXSettings settings:
    """
    mono_template = False
    if len(new_keys) <= CFN_MAX_OUTPUTS:
        mono_template = True

    for key in new_keys:
        key.stack = xstack
        key.define_kms_key()
        if key and key.cfn_resource:
            key.init_outputs()
            key.generate_outputs()
            if mono_template:
                template.add_resource(key.cfn_resource)
                key.handle_key_settings(template)
                template.add_output(key.outputs)
            elif not mono_template:
                key_template = build_template(
                    f"Template for KMS key {key.logical_name}"
                )
                key_template.add_resource(key.cfn_resource)
                key.handle_key_settings(key_template)
                key_template.add_output(key.outputs)
                key_stack = ComposeXStack(key.logical_name, stack_template=key_template)
                template.add_resource(key_stack)



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

from os import path

from ecs_composex.common.cfn_params import Parameter
from ecs_composex.common.ecs_composex import X_KEY
from ecs_composex.vpc.vpc_params import SG_ID_TYPE

MOD_KEY = path.basename(path.dirname(path.abspath(__file__)))
RES_KEY = f"{X_KEY}{MOD_KEY}"

LB_ID_T = "elbv2Id"
LB_ARN_T = "elbv2Arn"
LB_SG_ID_T = "elbv2SecurityGroupId"
LB_DNS_NAME_T = "DNSName"
LB_DNS_ZONE_ID_T = "CanonicalHostedZoneID"

LB_ID = Parameter(LB_ID_T, Type="String")
LB_ARN = Parameter(LB_ARN_T, Type="String")
LB_SG_ID = Parameter(LB_SG_ID_T, return_value="GroupId", Type=SG_ID_TYPE)
LB_DNS_NAME = Parameter(LB_DNS_NAME_T, return_value="DNSName", Type="String")
LB_DNS_ZONE_ID = Parameter(
    LB_DNS_ZONE_ID_T,
    return_value="CanonicalHostedZoneID",
    Type="String",
    AllowedPattern=r"^Z[0-9A-Z]+$",
)
LB_NAME_T = "LoadBalancerName"
LB_NAME = Parameter(LB_NAME_T, return_value="LoadBalancerName", Type="String")
LB_FULL_NAME_T = "LoadBalancerFullName"
LB_FULL_NAME = Parameter(
    LB_FULL_NAME_T, return_value="LoadBalancerFullName", Type="String"
)
LB_SGS_IDS_T = "LBSecurityGroups"
LB_SGS_IDS = Parameter(
    LB_SGS_IDS_T,
    return_value="SecurityGroups",
    Type="List<AWS::EC2::SecurityGroup::Id>",
)

TGT_GROUP_ARN_T = "TargetGroupArn"
TGT_GROUP_ARN = Parameter(TGT_GROUP_ARN_T, Type="String")

TGT_GROUP_NAME_T = "TargetGroupName"
TGT_GROUP_NAME = Parameter(
    TGT_GROUP_NAME_T, return_value="TargetGroupName", Type="String"
)



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

"""
Module to define the ComposeX Resources into a simple object to make it easier to navigate through.
"""

from copy import deepcopy
from re import sub

from compose_x_common.compose_x_common import keyisset, keypresent
from troposphere import AWS_STACK_NAME, Export, FindInMap, GetAtt, If, Output, Ref, Sub
from troposphere.ecs import Environment

from ecs_composex.common import LOG, NONALPHANUM
from ecs_composex.common.cfn_conditions import USE_STACK_NAME_CON_T
from ecs_composex.common.cfn_params import ROOT_STACK_NAME, Parameter
from ecs_composex.common.ecs_composex import CFN_EXPORT_DELIMITER as DELIM
from ecs_composex.common.ecs_composex import X_KEY


def set_resources(settings, resource_class, res_key, mod_key=None, mapping_key=None):
    """
    Method to define the ComposeXResource for each service.

    :param ecs_composex.common.settings.ComposeXSettings settings:
    :param ecs_composex.common.compose_resources.XResource.__init__ resource_class:
    :param str res_key: The compose key identifier for resource
    :param str mod_key: The module name in ecs_composex mapping the resource type
    :param str mapping_key: The value of the Mapping map name for FindInMap
    """
    if not mod_key:
        mod_key = sub(X_KEY, "", res_key)
    if not keyisset(res_key, settings.compose_content):
        return
    for resource_name in settings.compose_content[res_key]:
        new_definition = resource_class(
            name=resource_name,
            definition=settings.compose_content[res_key][resource_name],
            module_name=mod_key,
            settings=settings,
            mapping_key=mapping_key,
        )
        LOG.debug(type(new_definition))
        LOG.debug(new_definition.__dict__)
        settings.compose_content[res_key][resource_name] = new_definition


def get_parameter_settings(resource, parameter):
    """
    Function to define a set of values for the purpose of exposing resources settings from their stack to another.

    :param resource: The XResource we want to extract the outputs from
    :param parameter: The parameter we want to extract the outputs for
    :return: Ordered combination of settings
    :rtype: tuple
    """
    return (
        resource.attributes_outputs[parameter]["Name"],
        resource.attributes_outputs[parameter]["ImportParameter"],
        resource.attributes_outputs[parameter]["ImportValue"],
        parameter,
    )


def get_setting_key(name, settings_dict):
    if keyisset(name.title(), settings_dict):
        return name.title()
    return name


class XResource(object):
    """
    Class to represent each defined resource in the template

    :cvar str name: The name of the resource as defined in compose file
    :cvar dict definition: The definition of the resource as defined in compose file
    :cvar str logical_name: Name of the resource to use in CFN template as for export/import
    """

    policies_scaffolds = {}

    def __init__(self, name, definition, module_name, settings, mapping_key=None):
        """
        Init the class
        :param str name: Name of the resource in the template
        :param dict definition: The definition of the resource as-is
        """
        self.name = name
        self.module_name = module_name
        self.mapping_key = mapping_key
        if self.mapping_key is None:
            self.mapping_key = self.module_name
        self.definition = deepcopy(definition)
        self.env_names = []
        self.env_vars = []
        self.logical_name = NONALPHANUM.sub("", self.name)
        self.settings = (
            None
            if not keyisset("Settings", self.definition)
            else self.definition["Settings"]
        )
        self.use = (
            None if not keyisset("Use", self.definition) else self.definition["Use"]
        )
        self.lookup = (
            None
            if not keyisset("Lookup", self.definition)
            else self.definition["Lookup"]
        )
        if keyisset("Properties", self.definition) and not self.lookup:
            self.properties = self.definition["Properties"]
        elif not keyisset("Properties", self.definition) and keypresent(
            "Properties", self.definition
        ):
            self.properties = {}
        else:
            self.properties = None
        self.services = (
            []
            if not keyisset("Services", self.definition)
            else self.definition["Services"]
        )
        self.parameters = (
            {}
            if not keyisset("MacroParameters", self.definition)
            else self.definition["MacroParameters"]
        )
        self.cfn_resource = None
        self.output_properties = {}
        self.outputs = []
        self.attributes_outputs = {}
        self.families_targets = []
        self.families_scaling = []
        self.subnets_override = None
        self.is_nested = False
        self.stack = None
        self.init_env_names()
        self.ref_parameter = Parameter(self.logical_name, Type="String")
        self.set_services_targets(settings)
        self.set_services_scaling(settings)
        self.mappings = {}

    def __repr__(self):
        return self.logical_name

    def debug_families_targets(self):
        """
        Method to troubleshoot family and service mapping
        """
        for family in self.families_targets:
            LOG.debug(f"Mapped {family[0].name} to {self.name}.")
            if not family[1] and family[2]:
                LOG.debug(f"Applies to service {family[2]}")
            else:
                LOG.debug(f"Applies to all services of {family[0].name}")

    def handle_families_targets_expansion(self, service, settings):
        """
        Method to list all families and services that are targets of the resource.
        Allows to implement family and service level association to resource

        :param dict service: Service definition in compose file
        :param ecs_composex.common.settings.ComposeXSettings settings: Execution settings
        """
        name_key = get_setting_key("name", service)
        access_key = get_setting_key("access", service)
        the_service = [s for s in settings.services if s.name == service[name_key]][0]
        for family_name in the_service.families:
            family_name = NONALPHANUM.sub("", family_name)
            if family_name not in [f[0].name for f in self.families_targets]:
                self.families_targets.append(
                    (
                        settings.families[family_name],
                        False,
                        [the_service],
                        service[access_key],
                        service,
                    )
                )

    def set_services_targets(self, settings):
        """
        Method to map services and families targets of the services defined.
        TargetStructure:
        (family, family_wide, services[], access)

        :param ecs_composex.common.settings.ComposeXSettings settings:
        :return:
        """
        if not self.services:
            LOG.debug(f"No services defined for {self.name}")
            return
        for service in self.services:
            name_key = get_setting_key("name", service)
            access_key = get_setting_key("access", service)
            service_name = service[name_key]
            if service_name in settings.families and service_name not in [
                f[0].name for f in self.families_targets
            ]:
                self.families_targets.append(
                    (
                        settings.families[service_name],
                        True,
                        settings.families[service_name].services,
                        service[access_key],
                        service,
                    )
                )
            elif service_name in settings.families and service_name in [
                f[0].name for f in self.families_targets
            ]:
                LOG.warning(
                    f"The family {service_name} has already been added. Skipping"
                )
            elif service_name in [s.name for s in settings.services]:
                self.handle_families_targets_expansion(service, settings)
        self.debug_families_targets()

    def handle_family_scaling_expansion(self, service, settings):
        """
        Method to search for the families of given service and add it if not already present

        :param dict service:
        :param ecs_composex.common.settings.ComposeXSettings settings:
        :return:
        """
        name_key = get_setting_key("name", service)
        scaling_key = get_setting_key("scaling", service)
        the_service = [s for s in settings.services if s.name == service[name_key]][0]
        for family_name in the_service.families:
            family_name = NONALPHANUM.sub("", family_name)
            if family_name not in [f[0].name for f in self.families_scaling]:
                self.families_scaling.append(
                    (settings.families[family_name], service[scaling_key])
                )

    def set_services_scaling(self, settings):
        """
        Method to map services and families targets of the services defined.

        :param ecs_composex.common.settings.ComposeXSettings settings:
        :return:
        """
        if not self.services:
            return
        for service in self.services:
            name_key = get_setting_key("name", service)
            scaling_key = get_setting_key("scaling", service)
            if not keyisset(scaling_key, service):
                LOG.debug(
                    f"No scaling for {service[name_key]} defined based on {self.name}"
                )
                continue
            service_name = service[name_key]
            if service_name in settings.families and service_name not in [
                f[0].name for f in self.families_scaling
            ]:
                self.families_scaling.append(
                    (settings.families[service_name], service[scaling_key])
                )
            elif service_name in settings.families and service_name in [
                f[0].name for f in self.families_scaling
            ]:
                LOG.debug(f"The family {service_name} has already been added. Skipping")
            elif service_name in [s.name for s in settings.services]:
                self.handle_family_scaling_expansion(service, settings)

    def init_env_names(self, add_self_default=True):
        """
        Method to define the environment variables for the resource

        :return: list of environment variable names
        :rtype: list
        """
        if add_self_default:
            self.env_names.append(self.name.replace("-", "_"))
        if (
            self.settings
            and keyisset("EnvNames", self.settings)
            and isinstance(self.settings["EnvNames"], list)
        ):
            for env_name in self.settings["EnvNames"]:
                if isinstance(env_name, str) and env_name not in self.env_names:
                    self.env_names.append(env_name)

    def define_ref_env_vars(self, env_name, parameter):
        """
        Method to define construct parameters for Environment Variable for default Ref value of resource

        :param str env_name:
        :param ecs_composex.common.cfn_params.Parameter parameter:
        :return: dict with the Name and Value for environment variable
        :rtype: dict
        """
        container_env_name = env_name
        if self.lookup:
            container_env_value = self.attributes_outputs[parameter]["ImportValue"]
        else:
            container_env_value = Ref(
                self.attributes_outputs[parameter]["ImportParameter"]
            )
        return {"Name": container_env_name, "Value": container_env_value}

    def define_return_value_env_vars(self, env_name, parameter):
        """
        Method to define construct parameters for Environment Variable for parameters with specific return_value

        :param str env_name:
        :param ecs_composex.common.cfn_params.Parameter parameter:
        :return: dict with the Name and Value for environment variable
        :rtype: dict
        """
        container_env_name = f"{env_name}_{parameter.return_value}"
        if self.lookup:
            container_env_value = Sub(
                f"${{ResourceName}}_{parameter.return_value}",
                ResourceName=self.attributes_outputs[parameter]["ImportValue"],
            )
        else:
            container_env_value = Ref(
                self.attributes_outputs[parameter]["ImportParameter"]
            )
        return {"Name": container_env_name, "Value": container_env_value}

    def generate_resource_envvars(self):
        """
        Method to define all the env var of a resource based on its own defined output attributes
        """
        for env_name in self.env_names:
            if self.cfn_resource:
                for parameter in self.output_properties.keys():
                    if parameter.return_value:
                        env_var = Environment(
                            **self.define_return_value_env_vars(env_name, parameter)
                        )
                    else:
                        env_var = Environment(
                            **self.define_ref_env_vars(env_name, parameter)
                        )
                    self.env_vars.append(env_var)
            elif not self.cfn_resource and self.mappings:
                for key in self.mappings.keys():
                    env_var = Environment(
                        Name=env_name
                        if key == self.logical_name
                        else f"{env_name}_{key}",
                        Value=FindInMap(self.mapping_key, self.logical_name, key),
                    )
                    self.env_vars.append(env_var)
        self.env_vars = list({v.Name: v for v in self.env_vars}.values())

    def set_attributes_from_mapping(self, attribute_parameter):
        """
        Method to define the attribute outputs for lookup resources, which use FindInMap or Ref

        :param attribute_parameter: The parameter mapped to the resource attribute
        :type attribute_parameter: ecs_composex.common.cfn_params.Parameter
        :return: The FindInMap setting for mapped resource
        """
        if attribute_parameter.return_value:
            return FindInMap(
                self.mapping_key,
                self.logical_name,
                attribute_parameter.return_value,
            )
        else:
            return FindInMap(
                self.mapping_key, self.logical_name, attribute_parameter.title
            )

    def define_export_name(self, output_definition, attribute_parameter):
        """
        Method to define the export name for the resource
        :return:
        """
        if len(output_definition) == 5 and output_definition[4]:
            LOG.debug(f"Adding portback output for {self.name}")
            export = Export(
                If(
                    USE_STACK_NAME_CON_T,
                    Sub(
                        f"${{{AWS_STACK_NAME}}}{DELIM}{self.name}{DELIM}{output_definition[4]}"
                    ),
                    Sub(
                        f"${{{ROOT_STACK_NAME.title}}}{DELIM}{self.name}{DELIM}{output_definition[4]}"
                    ),
                )
            )
        else:
            export = Export(
                If(
                    USE_STACK_NAME_CON_T,
                    Sub(
                        f"${{{AWS_STACK_NAME}}}{DELIM}{self.logical_name}{DELIM}{attribute_parameter.title}"
                    ),
                    Sub(
                        f"${{{ROOT_STACK_NAME.title}}}{DELIM}{self.logical_name}{DELIM}{attribute_parameter.title}"
                    ),
                )
            )
        return export

    def set_new_resource_outputs(self, output_definition, attribute_parameter):
        """
        Method to define the outputs for the resource when new
        """
        if output_definition[2] is Ref:
            value = Ref(output_definition[1])
        elif output_definition[2] is GetAtt:
            value = GetAtt(output_definition[1], output_definition[3])
        elif output_definition[2] is Sub:
            value = Sub(output_definition[3])
        else:
            raise TypeError(
                f"3rd argument for {output_definition[0]} must be one of",
                (Ref, GetAtt, Sub),
                "Got",
                output_definition[2],
            )
        export = self.define_export_name(output_definition, attribute_parameter)
        return value, export

    def generate_outputs(self):
        """
        Method to create the outputs for XResources
        """
        if self.stack and not self.stack.is_void:
            root_stack = self.stack.title
        else:
            root_stack = self.mapping_key
        for (
            attribute_parameter,
            output_definition,
        ) in self.output_properties.items():
            output_name = f"{self.logical_name}{attribute_parameter.title}"
            if self.lookup:
                self.attributes_outputs[attribute_parameter] = {
                    "Name": output_name,
                    "ImportValue": self.set_attributes_from_mapping(
                        attribute_parameter
                    ),
                    "ImportParameter": None,
                }
            else:
                settings = self.set_new_resource_outputs(
                    output_definition, attribute_parameter
                )
                value = settings[0]
                export = settings[1]
                self.attributes_outputs[attribute_parameter] = {
                    "Name": output_name,
                    "Output": Output(output_name, Value=value, Export=export),
                    "ImportParameter": Parameter(
                        output_name,
                        return_value=attribute_parameter.return_value,
                        Type=attribute_parameter.Type,
                    ),
                    "ImportValue": GetAtt(
                        root_stack,
                        f"Outputs.{output_name}",
                    ),
                    "Original": attribute_parameter,
                }
        for attr in self.attributes_outputs.values():
            if keyisset("Output", attr):
                self.outputs.append(attr["Output"])

    def set_override_subnets(self):
        if (
            self.settings
            and keyisset("Subnets", self.settings)
            and hasattr(self, "subnets_param")
        ):
            self.subnets_override = self.settings["Subnets"]



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

"""
Module to handle ECS services access.
"""

ACCESS_TYPES = {"Publish": {"Action": ["sns:Publish"], "Effect": "Allow"}}



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>


import re

from troposphere import Join, Ref, Sub
from troposphere.iam import Role

from ecs_composex.common import LOG

ROLE_ARN_ARG = "RoleArn"


def service_role_trust_policy(service_name):
    """
    Simple function to format the trust relationship for a Role and an AWS Service
    used from lambda-my-aws/ozone

    :param service_name: name of the ecs_service
    :type service_name: str

    :return: policy document
    :rtype: dict
    """
    statement = {
        "Effect": "Allow",
        "Principal": {"Service": [Sub(f"{service_name}.${{AWS::URLSuffix}}")]},
        "Action": ["sts:AssumeRole"],
        "Condition": {"Bool": {"aws:SecureTransport": "true"}},
    }
    policy_doc = {"Version": "2012-10-17", "Statement": [statement]}
    return policy_doc


def define_iam_policy(policy):
    policy_def = policy
    policy_re = re.compile(
        r"((^([a-zA-Z0-9-_./]+)$)|(^(arn:aws:iam::(aws|\d{12}):policy/)[a-zA-Z0-9-_./]+$))"
    )

    if not policy_re.match(policy):
        raise ValueError(
            f"policy name {policy} does not match expected regexp",
            policy_re.pattern,
        )
    if isinstance(policy, str) and not policy.startswith("arn:aws:iam::"):
        policy_def = Sub(
            f"arn:${{AWS::Partition}}:iam::${{AWS::AccountId}}:policy/{policy}"
        )
    elif isinstance(policy, (Sub, Ref, Join)):
        LOG.debug(f"policy {policy}")
    return policy_def


def add_role_boundaries(iam_role, policy):
    """
    Function to set permission boundary onto an IAM role

    :param iam_role: the IAM Role to add the boundary to
    :type iam_role: troposphere.iam.Role
    :param policy: the name or ARN of the policy
    :type policy: str
    """
    if not isinstance(iam_role, Role):
        raise TypeError(f"{iam_role} is of type", type(iam_role), "expected", Role)
    if isinstance(policy, str):
        policy = define_iam_policy(policy)
    if hasattr(iam_role, "PermissionsBoundary"):
        LOG.warning(
            f"IAM Role {iam_role.title} already has PermissionsBoundary set. Overriding"
        )
    setattr(iam_role, "PermissionsBoundary", policy)
